from __future__ import annotations

import filecmp
import re
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Union

import numpy as np
import pytest

from ropt.config.plan import PlanConfig
from ropt.enums import EventType, OptimizerExitCode
from ropt.plan import (
    BasicOptimizer,
    Event,
    OptimizerContext,
    Plan,
)
from ropt.report import ResultsTable
from ropt.results import FunctionResults, Results

if TYPE_CHECKING:
    from pathlib import Path


# ruff: noqa: SLF001


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "optimizer": {
            "tolerance": 1e-5,
            "max_functions": 20,
        },
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "objective_functions": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def test_run_basic(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
            "results": None,
        },
        "steps": [
            {
                "optimizer": {"config": "$enopt_config", "tags": "opt"},
            },
        ],
        "results": [
            {
                "tracker": {"var": "results", "tags": "opt"},
            },
        ],
    }
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    assert plan["plan_id"] == [0]
    plan.run()
    assert plan.plan_id == (0,)
    variables = plan["results"].evaluations.variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        ("", ""),
        (" ", ""),
        ("$(1)", 1),
        ("$('1')", "1"),
        (" $(-1) ", -1),
        ("$(not 1)", False),
        ("$(True and False)", False),
        ("$(True or False)", True),
        ("$(1 + 1)", 2),
        ("$(2**3)", 8),
        ("$(3 % 2)", 1),
        ("$(3 // 2)", 1),
        ("$(2.5 + (2 + 3) / 2)", 5),
        ("$(1 < 2)", True),
        ("$(1 < 23)", True),
        ("$(1 < 2 > 3)", False),
        ("$(x)", 1),
        ("$('x')", "x"),
        ("$('$x')", "$x"),
        ("$(x + y)", 2),
        ("$( [1, 2] + [0] )", [1, 2, 0]),
        ("$( [1, 2] + plan_id )", [1, 2, 0]),
        ("$( x + 1 )", 2),
        ("$( max(x + 1, 2) )", 2),
        ("$max(x + 1, 2)", 2),
        ("1 + $x", "1 + $x"),
        ("$(1 + y)", 2),
        ("$( {'a': {'b': 1}} )", {"a": {"b": 1}}),
        ("$dummy", None),
        ("$(dummy)", None),
        ("$$dummy", "$dummy"),
        ("$([1, 2])", [1, 2]),
        ("$([dummy, 2])", [None, 2]),
        ("$(max (1, 2))", 2),
        ("$(max (x, 2))", 2),
        ("$max (x, 2)", 2),
        ("$(incr(x))", 2),
        ("$incr(x)", 2),
        ("$incr(incr(x) + 1)", 4),
        ("[[1]]", "1"),
        ("[[[1]]]", "[1]"),
        ("[[<< x >>]]", "1"),
        ("[[ <<not 1>> ]]", " False "),
        ("[[<<True and False>>]]", "False"),
        ("[[a << 1 >> b]]", "a 1 b"),
        ("[[a << 1 + 1 >> b]]", "a 2 b"),
        ("[[<< {'a': {'b': 1}} >>]]", "{'a': {'b': 1}}"),
        (
            "[[<<  {'a': {i: 1}}  >> - << {'a': {'b': x}} >>]]",
            "{'a': {'b': 1}} - {'a': {'b': 1}}",
        ),
    ],
)
def test_eval_expr(evaluator: Any, expr: str, expected: Any) -> None:
    plan_config: Dict[str, Any] = {
        "variables": {
            "dummy": None,
            "x": 1,
            "y": 1,
            "i": "b",
        },
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    context.expr.add_functions({"incr": lambda x: x + 1})
    plan = Plan(parsed_config, context)
    plan.run()

    assert plan.eval(expr) == expected


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        ("$mkdict(keys, [3, 4])", {"a": 3, "b": 4}),
    ],
)
def test_eval_expr_functions(evaluator: Any, expr: str, expected: Any) -> None:
    plan_config: Dict[str, Any] = {
        "variables": {
            "keys": ["a", "b"],
        },
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()
    assert plan.eval(expr) == expected


@pytest.mark.parametrize(
    ("expr", "message", "exception"),
    [
        ("$(1 + * 1)", "invalid syntax", SyntaxError),
        ("$z", "Unknown plan variable: `z`", AttributeError),
        ("[[<< $(1 + * 1) >>]]", "invalid syntax", SyntaxError),
        ("[[<< $(1 + 1) )]]", re.escape("Missing, or too many, `>>`"), SyntaxError),
        (
            "[[<< $(1 + 1) >> >>]]",
            re.escape("Missing, or too many, `>>`"),
            SyntaxError,
        ),
        ("$(foo())", "Unknown plan function: `foo`", AttributeError),
    ],
)
def test_eval_exception(
    evaluator: Any,
    expr: str,
    message: Any,
    exception: Union[AttributeError, SyntaxError],
) -> None:
    plan_config: Dict[str, Any] = {}
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    context.expr.add_functions({"incr": lambda x: x + 1})
    plan = Plan(parsed_config, context)
    plan.run()
    with pytest.raises(exception, match=message):  # type: ignore[call-overload]
        plan.eval(expr)


def test_eval_attribute(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    plan_config: Dict[str, Any] = {
        "variables": {
            "config": enopt_config,
            "results": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$config",
                    "tags": "opt",
                },
            },
        ],
        "results": [
            {"tracker": {"var": "results", "tags": "opt"}},
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()
    assert isinstance(plan.eval("$results"), Results)
    assert plan.eval("$results.result_id") >= 0
    assert plan.eval("$(results.result_id)") >= 0
    assert plan.eval("$results.plan_id[0]") == 0
    assert plan.eval("$(results.plan_id[0])") == 0


@pytest.mark.parametrize(
    ("variables", "expr", "check", "expected"),
    [
        ({"x": None}, {"x": 1}, "x", 1),
        ({"x": None, "y": None}, [{"x": 1}, {"y": "$(x + 1)"}], "y", 2),
        ({"x": {"a": 1}, "i": "a"}, {"x[i]": 2}, "x", {"a": 2}),
        ({"x": {"a": {10: {}}}, "i": 5}, {"x['a'][i + 5]": 1}, "x", {"a": {10: 1}}),
    ],
)
def test_set_step(
    evaluator: Any,
    variables: Dict[str, Any],
    expr: Dict[str, Any],
    check: str,
    expected: Any,
) -> None:
    plan_config: Dict[str, Any] = {
        "variables": variables,
        "steps": [{"set": expr}],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()
    assert plan[check] == expected


@pytest.mark.parametrize(
    ("variables", "expr", "check", "expected"),
    [
        (
            {"o": type("Obj", (object,), {"b_b": type("Obj", (object,), {"c1": 1})})},
            {"o.b_b.c1": 2},
            'plan["o"].b_b.c1',
            2,
        ),
        (
            {
                "o": type(
                    "Obj", (object,), {"b_b": type("Obj", (object,), {"c1": {20: 1}})}
                )
            },
            {"o.b_b.c1[20]": 2},
            'plan["o"].b_b.c1',
            {20: 2},
        ),
        (
            {"o": type("Obj", (object,), {"b_b": [type("Obj", (object,), {"c1": 1})]})},
            {"o.b_b[0].c1": 2},
            'plan["o"].b_b[0].c1',
            2,
        ),
    ],
)
def test_set_step_attribute(
    evaluator: Any,
    variables: Dict[str, Any],
    expr: Dict[str, Any],
    check: str,
    expected: Any,
) -> None:
    plan_config: Dict[str, Any] = {
        "variables": variables,
        "steps": [{"set": expr}],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()
    assert eval(check) == expected  # noqa: S307


def test_set_keys_exception(evaluator: Any) -> None:
    plan_config: Dict[str, Any] = {
        "variables": {
            "y": {"a": None},
        },
        "steps": [
            {"set": {"y['a']['b']": 1}},
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    with pytest.raises(
        AttributeError, match=re.escape("Invalid attribute access: y['a']['b']")
    ):
        plan.run()


def test_set_keys_value_exception(evaluator: Any) -> None:
    plan_config: Dict[str, Any] = {
        "variables": {
            "y": {"a": None},
        },
        "steps": [
            {"set": {"y['a']": "$foo(1)"}},
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    with pytest.raises(AttributeError, match=re.escape("Unknown plan function: `foo`")):
        plan.run()


@pytest.mark.parametrize("expr1", ["$(1 > 0)", "1 > 0"])
@pytest.mark.parametrize("expr2", ["$(x < 0)", "x < 0"])
def test_conditional_run(
    enopt_config: Dict[str, Any], evaluator: Any, expr1: str, expr2: str
) -> None:
    plan_config = {
        "variables": {
            "config": enopt_config,
            "optimal1": None,
            "optimal2": None,
            "x": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$config",
                    "tags": "optimal1",
                },
                "if": expr1,
            },
            {"set": {"x": 1}},
            {
                "if": expr2,
                "optimizer": {
                    "config": "$config",
                    "tags": "optimal2",
                },
            },
        ],
        "results": [
            {
                "tracker": {"var": "optimal1", "tags": "optimal1"},
            },
            {
                "tracker": {"var": "optimal2", "tags": "optimal2"},
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()
    result1 = plan["optimal1"]
    result2 = plan["optimal2"]
    assert result1 is not None
    assert np.allclose(result1.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    assert result2 is None


def test_plan_rng(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    enopt_config["gradient"]["seed"] = 1

    plan_config = {
        "variables": {
            "config": enopt_config,
            "optimal1": None,
            "optimal2": None,
            "optimal3": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$config",
                    "tags": "optimal1",
                },
            },
            {
                "optimizer": {
                    "config": "$config",
                    "tags": "optimal2",
                },
            },
            {
                "set": {
                    "config['gradient']['seed']": "$(plan_id + [config['gradient']['seed']])"
                }
            },
            {
                "optimizer": {
                    "config": "$config",
                    "tags": "optimal3",
                },
            },
        ],
        "results": [
            {
                "tracker": {"var": "optimal1", "tags": ["optimal1"]},
            },
            {
                "tracker": {"var": "optimal2", "tags": ["optimal2"]},
            },
            {
                "tracker": {"var": "optimal3", "tags": ["optimal3"]},
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    result1 = plan["optimal1"]
    assert result1 is not None
    assert np.allclose(result1.evaluations.variables, [0.0, 0.0, 0.5], atol=0.025)

    result2 = plan["optimal2"]
    assert result2 is not None
    assert np.allclose(result2.evaluations.variables, [0.0, 0.0, 0.5], atol=0.025)

    result3 = plan["optimal3"]
    assert result3 is not None
    assert np.allclose(result3.evaluations.variables, [0.0, 0.0, 0.5], atol=0.025)

    assert np.all(result1.evaluations.variables == result2.evaluations.variables)
    assert not np.all(result1.evaluations.variables == result3.evaluations.variables)


def test_set_initial_values(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    plan_config = {
        "variables": {
            "config": enopt_config,
            "optimal1": None,
            "optimal2": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$config",
                    "tags": "optimal1",
                },
            },
            {
                "optimizer": {
                    "config": "$config",
                    "tags": "optimal2",
                    "initial_values": [0, 0, 0],
                },
            },
        ],
        "results": [
            {
                "tracker": {"var": "optimal1", "tags": ["optimal1"]},
            },
            {
                "tracker": {"var": "optimal2", "tags": ["optimal2"]},
            },
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    result1 = plan["optimal1"]
    assert result1 is not None
    assert np.allclose(result1.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)

    result2 = plan["optimal2"]
    assert result2 is not None
    assert np.allclose(result2.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)

    assert not np.all(result1.evaluations.variables == result2.evaluations.variables)


def test_reset_results(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    plan_config = {
        "variables": {
            "config": enopt_config,
            "optimal": None,
            "saved_results": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$config",
                    "tags": "opt",
                },
            },
            {
                "set": {
                    "saved_results": "$optimal",
                    "optimal": None,
                },
            },
        ],
        "results": [
            {"tracker": {"var": "optimal", "tags": "opt"}},
        ],
    }
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    assert plan["optimal"] is None
    saved_results = plan["saved_results"]
    assert saved_results is not None
    assert np.allclose(saved_results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)


def test_two_optimizers_alternating(
    enopt_config: Dict[str, Any], evaluator: Any
) -> None:
    completed_functions = 0

    def _track_evaluations(event: Event) -> None:
        nonlocal completed_functions
        assert event.results is not None
        for item in event.results:
            if isinstance(item, FunctionResults):
                completed_functions += 1

    enopt_config["variables"]["initial_values"] = [0.0, 0.2, 0.1]

    opt_config1 = {
        "speculative": True,
        "max_functions": 4,
    }
    opt_config2 = {
        "speculative": True,
        "max_functions": 3,
    }

    enopt_config1 = deepcopy(enopt_config)
    enopt_config1["variables"]["indices"] = [0, 2]
    enopt_config1["optimizer"] = opt_config1
    enopt_config2 = deepcopy(enopt_config)
    enopt_config2["variables"]["indices"] = [1]
    enopt_config2["optimizer"] = opt_config2

    plan_config = {
        "variables": {
            "enopt_config1": enopt_config1,
            "enopt_config2": enopt_config2,
            "optimum": None,
            "last": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$enopt_config1",
                    "tags": "opt",
                },
            },
            {
                "optimizer": {
                    "config": "$enopt_config2",
                    "initial_values": "$last",
                    "tags": "opt",
                },
            },
            {
                "optimizer": {
                    "config": "$enopt_config1",
                    "initial_values": "$last",
                    "tags": "opt",
                },
            },
            {
                "optimizer": {
                    "config": "$enopt_config2",
                    "initial_values": "$last",
                    "tags": "opt",
                },
            },
        ],
        "results": [
            {"tracker": {"var": "optimum", "tags": "opt"}},
            {
                "tracker": {"var": "last", "type": "last", "tags": "opt"},
            },
        ],
    }

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()

    assert completed_functions == 14
    assert plan["optimum"] is not None
    assert np.allclose(
        plan["optimum"].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02
    )


def test_optimization_sequential(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    completed: List[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 2

    enopt_config2 = deepcopy(enopt_config)
    enopt_config2["optimizer"]["max_functions"] = 3

    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
            "enopt_config2": enopt_config2,
            "last": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$enopt_config",
                    "tags": "last",
                },
            },
            {
                "optimizer": {
                    "config": "$enopt_config2",
                    "initial_values": "$last",
                },
            },
        ],
        "results": [
            {
                "tracker": {"var": "last", "type": "last", "tags": "last"},
            },
        ],
    }

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()

    assert not np.allclose(
        completed[1].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02
    )
    assert np.all(
        completed[2].evaluations.variables == completed[1].evaluations.variables
    )
    assert np.allclose(completed[-1].evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)


def test_repeat_step(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
            "optimum": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$enopt_config",
                    "tags": "opt",
                },
            },
        ],
        "results": [
            {"tracker": {"var": "optimum", "tags": "opt"}},
        ],
    }
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()
    assert plan["optimum"] is not None
    variables = plan["optimum"].evaluations.variables.copy()
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.025)

    plan_config["steps"] = [
        {
            "repeat": {
                "iterations": 1,
                "steps": [
                    {
                        "optimizer": {
                            "config": "$enopt_config",
                            "tags": "opt",
                        },
                    },
                ],
            },
        }
    ]
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()
    assert plan["optimum"] is not None

    assert np.all(variables == plan["optimum"].evaluations.variables)


def test_restart_initial(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    completed: List[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 3

    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
        },
        "steps": [
            {
                "repeat": {
                    "iterations": 2,
                    "steps": [
                        {
                            "optimizer": {
                                "config": "$enopt_config",
                            },
                        },
                    ],
                },
            }
        ],
    }

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()

    assert len(completed) == 6

    initial = np.array([enopt_config["variables"]["initial_values"]])
    assert np.all(completed[0].evaluations.variables == initial)
    assert np.all(completed[3].evaluations.variables == initial)


def test_restart_last(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    completed: List[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 3

    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
            "last": None,
        },
        "steps": [
            {
                "repeat": {
                    "iterations": 2,
                    "steps": [
                        {
                            "optimizer": {
                                "config": "$enopt_config",
                                "initial_values": "$last",
                                "tags": "opt",
                            },
                        },
                    ],
                },
            }
        ],
        "results": [
            {
                "tracker": {"var": "last", "type": "last", "tags": "opt"},
            },
        ],
    }
    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()

    assert np.all(
        completed[3].evaluations.variables == completed[2].evaluations.variables
    )


def test_restart_optimum(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    completed: List[FunctionResults] = []

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
            "optimum": None,
        },
        "steps": [
            {
                "repeat": {
                    "iterations": 2,
                    "steps": [
                        {
                            "optimizer": {
                                "config": "$enopt_config",
                                "tags": "opt",
                                "initial_values": "$optimum",
                            },
                        },
                    ],
                },
            }
        ],
        "results": [
            {"tracker": {"var": "optimum", "tags": "opt"}},
        ],
    }
    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()

    assert np.all(
        completed[2].evaluations.variables == completed[4].evaluations.variables
    )


def test_restart_optimum_with_reset(
    enopt_config: Dict[str, Any], evaluator: Any, test_functions: Any
) -> None:
    completed: List[FunctionResults] = []
    max_functions = 5

    def _track_evaluations(event: Event) -> None:
        nonlocal completed
        assert event.results is not None
        completed += [
            item for item in event.results if isinstance(item, FunctionResults)
        ]

    # Make sure each restart has worse objectives, and that the last evaluation
    # is even worse, so each run has its own optimum that is worse than the
    # global and not at its last evaluation. We should end up with initial
    # values that are not from the global optimum, or from the most recent
    # evaluation:
    new_functions = (
        lambda variables, context: (
            test_functions[0](variables, context)
            + int((len(completed) + 1) / max_functions)
        ),
        lambda variables, context: (
            test_functions[1](variables, context)
            + int((len(completed) + 1) / max_functions)
        ),
    )

    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = max_functions

    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
            "optimum": None,
            "initial": None,
        },
        "steps": [
            {
                "repeat": {
                    "iterations": 3,
                    "steps": [
                        {
                            "set": {
                                "initial": "$optimum",
                                "optimum": None,
                            },
                        },
                        {
                            "optimizer": {
                                "config": "$enopt_config",
                                "initial_values": "$initial",
                                "tags": "opt",
                            },
                        },
                    ],
                },
            },
        ],
        "results": [
            {"tracker": {"var": "optimum", "tags": "opt"}},
        ],
    }
    context = OptimizerContext(evaluator=evaluator(new_functions)).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()

    # The third evaluation is the optimum, and used to restart the second run:
    assert np.all(
        completed[max_functions].evaluations.variables
        == completed[2].evaluations.variables
    )
    # The 5th evaluation is the optimum of the second run, and used for the third:
    assert np.all(
        completed[2 * max_functions].evaluations.variables
        == completed[5].evaluations.variables
    )


def test_repeat_metadata(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    restarts: List[int] = []

    def _track_results(event: Event) -> None:
        assert event.results is not None
        metadata = event.results[0].metadata
        restart = metadata.get("restart", -1)
        assert metadata["foo"] == 1
        assert metadata["bar"] == "string"
        if "complex" in metadata:
            assert metadata["complex"] == f"string 2 {restart}"
        if not restarts or restart != restarts[-1]:
            restarts.append(restart)

    metadata = {
        "restart": "$counter",
        "foo": 1,
        "bar": "string",
        "complex": "[[string <<1 + 1>> <<counter>>]]",
    }

    plan_config = {
        "variables": {
            "config": enopt_config,
            "metadata": metadata,
            "counter": 0,
        },
        "steps": [
            {
                "repeat": {
                    "iterations": 2,
                    "var": "counter",
                    "steps": [
                        {
                            "optimizer": {"config": "$config", "tags": "opt"},
                        },
                    ],
                },
            }
        ],
        "results": [
            {"metadata": {"data": metadata, "tags": "opt"}},
        ],
    }

    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_results
    )
    plan = Plan(parsed_config, context)
    plan.run()
    assert restarts == [0, 1]


def test_evaluator_step(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    plan_config: Dict[str, Any] = {
        "variables": {
            "config": enopt_config,
            "result": None,
        },
        "steps": [
            {
                "evaluator": {
                    "config": "$config",
                    "tags": "eval",
                },
            },
        ],
        "results": [
            {"tracker": {"var": "result", "tags": "eval"}},
        ],
    }

    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    assert plan["result"] is not None
    assert plan["result"].functions is not None
    assert np.allclose(plan["result"].functions.weighted_objective, 1.66)

    plan_config["steps"][0]["evaluator"]["values"] = [0, 0, 0]
    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    assert plan["result"] is not None
    assert plan["result"].functions is not None
    assert np.allclose(plan["result"].functions.weighted_objective, 1.75)


def test_evaluator_step_multi(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    plan_config = {
        "variables": {
            "config": enopt_config,
            "results": None,
        },
        "steps": [
            {
                "evaluator": {
                    "config": "$config",
                    "values": [[0, 0, 0.1], [0, 0, 0]],
                    "tags": "eval",
                },
            },
        ],
        "results": [
            {"tracker": {"var": "results", "type": "all", "tags": "eval"}},
        ],
    }

    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()
    values = [
        results.functions.weighted_objective.item() for results in plan["results"]
    ]
    assert np.allclose(values, [1.66, 1.75])


def test_exit_code(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4

    is_called = False

    def _exit_code(
        event: Event,
    ) -> None:
        nonlocal is_called
        is_called = True
        assert isinstance(event, Event)
        assert event.exit_code == OptimizerExitCode.MAX_FUNCTIONS_REACHED

    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
            "exit_code": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$enopt_config",
                    "exit_code_var": "exit_code",
                },
            },
        ],
    }
    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_OPTIMIZER_STEP, _exit_code
    )
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()
    assert plan["exit_code"] == OptimizerExitCode.MAX_FUNCTIONS_REACHED
    assert is_called


def test_nested_plan(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    enopt_config["variables"]["initial_values"] = [0.0, 0.2, 0.1]

    completed_functions = 0

    def _track_evaluations(event: Event) -> None:
        nonlocal completed_functions
        if "outer" in event.tags:
            assert event.plan_id == (0,)
        if "inner" in event.tags:
            assert event.plan_id == (0, completed_functions // 5)
        assert event.results is not None
        for item in event.results:
            if isinstance(item, FunctionResults):
                completed_functions += 1

    enopt_config["optimizer"]["tolerance"] = 1e-10
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4
    enopt_config["variables"]["indices"] = [0, 2]
    nested_config = deepcopy(enopt_config)
    nested_config["variables"]["indices"] = [1]
    enopt_config["optimizer"]["max_functions"] = 5

    inner_config = {
        "inputs": ["initial", "config"],
        "outputs": ["nested_optimum"],
        "steps": [
            {
                "optimizer": {
                    "config": "$config",
                    "initial_values": "$initial",
                    "tags": "inner",
                },
            },
        ],
        "results": [
            {
                "tracker": {"var": "nested_optimum", "tags": "inner"},
            },
        ],
    }

    outer_config = {
        "variables": {
            "config": enopt_config,
            "optimum": None,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$config",
                    "nested_optimization": {
                        "plan": inner_config,
                        "extra_inputs": [nested_config],
                    },
                    "tags": "outer",
                },
            },
        ],
        "results": [
            {
                "tracker": {"var": "optimum", "tags": "outer"},
            },
        ],
    }

    parsed_config = PlanConfig.model_validate(outer_config)
    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(parsed_config, context)
    plan.run()
    results = plan["optimum"]

    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    assert completed_functions == 25


def test_nested_plan_metadata(enopt_config: Dict[str, Any], evaluator: Any) -> None:
    enopt_config["variables"]["initial_values"] = [0.0, 0.2, 0.1]

    def _track_evaluations(event: Event) -> None:
        assert event.results is not None
        for item in event.results:
            if isinstance(item, FunctionResults):
                assert item.metadata.get("outer") == 1
                if "inner" in event.tags:
                    assert item.metadata.get("inner") == "inner_meta_data"

    enopt_config["optimizer"]["tolerance"] = 1e-10
    enopt_config["optimizer"]["speculative"] = True
    enopt_config["optimizer"]["max_functions"] = 4
    enopt_config["variables"]["indices"] = [0, 2]
    nested_config = deepcopy(enopt_config)
    nested_config["variables"]["indices"] = [1]
    enopt_config["optimizer"]["max_functions"] = 5

    inner_config = {
        "inputs": ["initial"],
        "outputs": ["nested_optimum"],
        "variables": {
            "config": nested_config,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$config",
                    "initial_values": "$initial",
                    "tags": "inner",
                },
            },
        ],
        "results": [
            {"tracker": {"var": "nested_optimum", "tags": "inner"}},
            {
                "metadata": {"data": {"inner": "inner_meta_data"}, "tags": "inner"},
            },
        ],
    }

    outer_config = {
        "variables": {
            "config": enopt_config,
            "optimum": None,
            "x": 1,
        },
        "steps": [
            {
                "optimizer": {
                    "config": "$config",
                    "nested_optimization": {"plan": inner_config},
                    "tags": "outer",
                },
            },
        ],
        "results": [
            {
                "metadata": {"data": {"outer": "$x"}, "tags": ["inner", "outer"]},
            },
            {"tracker": {"var": "optimum", "tags": "inner"}},
        ],
    }

    parsed_config = PlanConfig.model_validate(outer_config)
    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _track_evaluations
    )
    plan = Plan(parsed_config, context)
    plan.run()
    results = plan["optimum"]

    assert results.metadata["inner"] == "inner_meta_data"
    assert results.metadata["outer"] == 1

    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)


def test_table(enopt_config: Dict[str, Any], evaluator: Any, tmp_path: Path) -> None:
    enopt_config["optimizer"]["max_functions"] = 5

    path1 = tmp_path / "results1.txt"
    table = ResultsTable(
        columns={
            "result_id": "eval-ID",
            "evaluations.variables": "Variables",
        },
        path=path1,
    )
    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
        },
        "steps": [
            {
                "optimizer": {"config": "$enopt_config"},
            },
        ],
    }

    def handle_results(event: Event) -> None:
        assert event.results is not None
        table.add_results(event.config, event.results)

    context = OptimizerContext(evaluator=evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, handle_results
    )
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()

    assert path1.exists()
    with path1.open() as fp:
        assert len(fp.readlines()) == 8

    path2 = tmp_path / "results2.txt"
    BasicOptimizer(enopt_config, evaluator()).add_table(
        columns={"result_id": "eval-ID", "evaluations.variables": "Variables"},
        path=path2,
    ).run()
    assert path2.exists()

    assert filecmp.cmp(path1, path2)


def test_table_handler(
    enopt_config: Dict[str, Any], evaluator: Any, tmp_path: Path
) -> None:
    enopt_config["optimizer"]["max_functions"] = 5

    path1 = tmp_path / "results1.txt"
    plan_config = {
        "variables": {
            "enopt_config": enopt_config,
        },
        "steps": [
            {
                "optimizer": {"config": "$enopt_config", "tags": "opt"},
            },
        ],
        "results": [
            {
                "table": {
                    "tags": "opt",
                    "columns": {
                        "result_id": "eval-ID",
                        "evaluations.variables": "Variables",
                    },
                    "path": path1,
                },
            },
        ],
    }

    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()

    assert path1.exists()
    with path1.open() as fp:
        assert len(fp.readlines()) == 8

    path2 = tmp_path / "results2.txt"
    BasicOptimizer(enopt_config, evaluator()).add_table(
        columns={"result_id": "eval-ID", "evaluations.variables": "Variables"},
        path=path2,
    ).run()
    assert path2.exists()

    assert filecmp.cmp(path1, path2)


@pytest.mark.parametrize("file_format", ["json", "pickle"])
def test_save_step(evaluator: Any, file_format: str, tmp_path: Path) -> None:
    path = tmp_path / "vars"
    plan_config = {
        "variables": {
            "x": 1,
            "y": 2,
        },
        "steps": [
            {
                "save": {"vars": ["x", "y"], "path": path, "format": file_format},
            },
            {
                "set": {"x": None, "y": None},
            },
            {
                "load": {"vars": ["x", "y"], "path": path, "format": file_format},
            },
        ],
    }

    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(PlanConfig.model_validate(plan_config), context)
    plan.run()

    assert path.exists()
    assert plan["x"] == 1
    assert plan["y"] == 2


def test_save_results_step(
    enopt_config: Dict[str, Any], evaluator: Any, tmp_path: Path
) -> None:
    path = tmp_path / "results"
    plan_config: Dict[str, Any] = {
        "variables": {
            "config": enopt_config,
            "result": None,
        },
        "steps": [
            {
                "evaluator": {
                    "config": "$config",
                    "tags": "eval",
                },
            },
        ],
        "results": [
            {"tracker": {"var": "result", "tags": "eval", "type": "last"}},
            {"save": {"path": path, "tags": "eval"}},
        ],
    }

    parsed_config = PlanConfig.model_validate(plan_config)
    context = OptimizerContext(evaluator=evaluator())
    plan = Plan(parsed_config, context)
    plan.run()

    assert plan["result"] is not None
    assert plan["result"].functions is not None
    assert np.allclose(plan["result"].functions.weighted_objective, 1.66)

    assert path.with_suffix(".nc").exists()

    results = FunctionResults.from_netcdf(path.with_suffix(".nc"))
    assert isinstance(results, FunctionResults)
    assert results.functions is not None
    assert np.allclose(results.functions.weighted_objective, 1.66)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, cast

import numpy as np
import pytest
from numpy.typing import NDArray

from ropt.config.enopt import EnOptConfig
from ropt.config.enopt.constants import DEFAULT_SEED
from ropt.enums import ConstraintType, EventType, OptimizerExitCode
from ropt.plan import BasicOptimizer
from ropt.results import FunctionResults, GradientResults

if TYPE_CHECKING:
    from ropt.plan import Event


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "optimizer": {
            "tolerance": 1e-5,
            "max_functions": 20,
        },
        "variables": {
            "initial_values": [0.0, 0.0, 0.1],
        },
        "objective_functions": {
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "perturbation_magnitudes": 0.01,
        },
    }


def test_max_functions_exceeded(enopt_config: Any, evaluator: Any) -> None:
    last_evaluation = 100

    def track_results(event: Event) -> None:
        nonlocal last_evaluation
        assert event.results
        assert isinstance(event.results[0].result_id, int)
        last_evaluation = event.results[0].result_id

    max_functions = 2
    enopt_config["optimizer"]["max_functions"] = max_functions
    optimizer = BasicOptimizer(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, track_results
    )
    optimizer.run()
    assert last_evaluation == max_functions
    assert optimizer.exit_code == OptimizerExitCode.MAX_FUNCTIONS_REACHED


def test_max_functions_not_exceeded(enopt_config: Any, evaluator: Any) -> None:
    last_evaluation = 100

    def track_results(event: Event) -> None:
        nonlocal last_evaluation
        assert event.results
        assert isinstance(event.results[0].result_id, int)
        last_evaluation = event.results[0].result_id

    max_functions = 100
    enopt_config["optimizer"]["max_functions"] = max_functions
    enopt_config["optimizer"]["split_evaluations"] = True
    optimizer = BasicOptimizer(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, track_results
    )
    optimizer.run()
    assert last_evaluation + 1 < 2 * max_functions
    assert optimizer.exit_code == OptimizerExitCode.OPTIMIZER_STEP_FINISHED


def test_failed_realizations(enopt_config: Any, evaluator: Any) -> None:
    def _observer(event: Event) -> None:
        assert event.results
        assert isinstance(event.results[0], FunctionResults)
        assert event.results[0].functions is None

    functions = [lambda _0, _1: np.array(1.0), lambda _0, _1: np.array(np.nan)]
    optimizer = BasicOptimizer(enopt_config, evaluator(functions)).add_observer(
        EventType.FINISHED_EVALUATION, _observer
    )
    optimizer.run()
    assert optimizer.exit_code == OptimizerExitCode.TOO_FEW_REALIZATIONS


def test_all_failed_realizations_not_supported(
    enopt_config: Any, evaluator: Any
) -> None:
    enopt_config["realizations"] = {"realization_min_success": 0}

    functions = [lambda _0, _1: np.array(1.0), lambda _0, _1: np.array(np.nan)]
    optimizer = BasicOptimizer(
        enopt_config,
        evaluator(functions),
    )
    optimizer.run()
    assert optimizer.exit_code == OptimizerExitCode.TOO_FEW_REALIZATIONS


def test_user_abort(enopt_config: Any, evaluator: Any) -> None:
    last_evaluation = 100

    def _observer(event: Event) -> None:
        nonlocal last_evaluation
        assert event.results
        assert isinstance(event.results[0].result_id, int)
        last_evaluation = event.results[0].result_id
        if event.results[0].result_id == 1:
            optimizer.abort_optimization()

    optimizer = BasicOptimizer(enopt_config, evaluator()).add_observer(
        EventType.FINISHED_EVALUATION, _observer
    )
    optimizer.run()
    assert optimizer.results is not None
    assert last_evaluation == 1
    assert optimizer.exit_code == OptimizerExitCode.USER_ABORT


def test_single_perturbation(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["gradient"] = {
        "number_of_perturbations": 1,
        "merge_realizations": True,
    }
    enopt_config["realizations"] = {"weights": 5 * [1]}
    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)


def test_objective_auto_scale(
    enopt_config: Any, evaluator: Any, test_functions: Any
) -> None:
    config = EnOptConfig.model_validate(enopt_config)
    init1 = test_functions[1](config.variables.initial_values, None)

    enopt_config["objective_functions"]["scales"] = [1.0, init1]
    enopt_config["objective_functions"]["auto_scale"] = False
    manual_result = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert manual_result is not None

    enopt_config["objective_functions"]["scales"] = [1.0, 1.0]
    enopt_config["objective_functions"]["auto_scale"] = [False, True]
    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, manual_result)

    enopt_config["objective_functions"]["scales"] = [1.0, 2.0 * init1]
    enopt_config["objective_functions"]["auto_scale"] = False
    manual_result = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert manual_result is not None

    enopt_config["objective_functions"]["scales"] = [1.0, 2.0]
    enopt_config["objective_functions"]["auto_scale"] = [False, True]
    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, manual_result)


def test_constraint_auto_scale(
    enopt_config: Any, evaluator: Any, test_functions: Any
) -> None:
    enopt_config["nonlinear_constraints"] = {
        "rhs_values": 0.4,
        "types": ConstraintType.GE,
    }

    test_functions = (
        *test_functions,
        lambda variables, _: cast(NDArray[np.float64], variables[0] + variables[2]),
    )

    config = EnOptConfig.model_validate(enopt_config)
    scales = np.fabs(test_functions[-1](config.variables.initial_values, None))

    def check_constraints(event: Event) -> None:
        assert event.results
        for item in event.results:
            if isinstance(item, FunctionResults) and item.result_id == 0:
                assert item.functions is not None
                assert item.functions.scaled_constraints is not None
                assert np.allclose(item.functions.scaled_constraints, 1.0)

    enopt_config["nonlinear_constraints"]["scales"] = scales
    enopt_config["nonlinear_constraints"]["auto_scale"] = False
    BasicOptimizer(enopt_config, evaluator(test_functions)).add_observer(
        EventType.FINISHED_EVALUATION, check_constraints
    ).run()

    enopt_config["nonlinear_constraints"]["scales"] = 1.0
    enopt_config["nonlinear_constraints"]["auto_scale"] = True
    BasicOptimizer(enopt_config, evaluator(test_functions)).add_observer(
        EventType.FINISHED_EVALUATION, check_constraints
    ).run()


@pytest.mark.parametrize("offsets", [None, np.array([1.0, 1.1, 1.2])])
@pytest.mark.parametrize("scales", [None, np.array([2.0, 2.1, 2.2])])
def test_variables_scale(
    enopt_config: Any,
    evaluator: Any,
    offsets: Optional[NDArray[np.float64]],
    scales: Optional[NDArray[np.float64]],
) -> None:
    initial_values = np.array(enopt_config["variables"]["initial_values"])
    lower_bounds = np.array([-2.0, -np.inf, -3.0])
    upper_bounds = np.array([np.inf, 1.0, 4.0])

    enopt_config["optimizer"]["max_iterations"] = 20
    enopt_config["variables"]["lower_bounds"] = lower_bounds
    enopt_config["variables"]["upper_bounds"] = upper_bounds

    if offsets is not None:
        enopt_config["variables"]["offsets"] = offsets
    if scales is not None:
        enopt_config["variables"]["scales"] = scales

    results = BasicOptimizer(enopt_config, evaluator()).run().results
    assert results is not None

    if offsets is not None:
        initial_values = initial_values - offsets
        lower_bounds = lower_bounds - offsets
        upper_bounds = upper_bounds - offsets
    if scales is not None:
        initial_values = initial_values / scales
        lower_bounds = lower_bounds / scales
        upper_bounds = upper_bounds / scales
    config = EnOptConfig.model_validate(enopt_config)
    assert np.allclose(config.variables.initial_values, initial_values)
    assert np.allclose(config.variables.lower_bounds, lower_bounds)
    assert np.allclose(config.variables.upper_bounds, upper_bounds)
    result = results.evaluations.scaled_variables
    if result is None:
        result = results.evaluations.variables
    if scales is not None:
        result = result * scales
    if offsets is not None:
        result = result + offsets
    assert np.allclose(result, [0.0, 0.0, 0.5], atol=0.05)
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.05)


def test_variables_scale_linear_constraints(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 0, 1], [0, 1, 1]],
        "rhs_values": [1.0, 0.75],
        "types": [ConstraintType.EQ, ConstraintType.EQ],
    }

    offsets = np.array([1.0, 1.1, 1.2])
    scales = np.array([2.0, 2.1, 2.2])
    enopt_config["variables"]["offsets"] = offsets
    enopt_config["variables"]["scales"] = scales

    config = EnOptConfig.model_validate(enopt_config)
    assert config.linear_constraints is not None
    coefficients = config.linear_constraints.coefficients
    rhs_values = config.linear_constraints.rhs_values
    assert np.allclose(
        coefficients / scales, enopt_config["linear_constraints"]["coefficients"]
    )
    assert np.allclose(
        rhs_values
        + np.matmul(coefficients if scales is None else coefficients / scales, offsets),
        enopt_config["linear_constraints"]["rhs_values"],
    )

    results = BasicOptimizer(enopt_config, evaluator()).run().results
    assert results is not None
    assert results.evaluations.variables is not None
    assert np.allclose(results.evaluations.variables, [0.25, 0.0, 0.75], atol=0.02)


def test_check_linear_constraints(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 1, 0], [1, 1, 0], [1, 1, 0]],
        "rhs_values": [0.0, 1.0, -1.0],
        "types": [ConstraintType.EQ, ConstraintType.LE, ConstraintType.GE],
    }
    enopt_config["optimizer"]["max_functions"] = 1
    results = BasicOptimizer(enopt_config, evaluator()).run().results
    assert results is not None

    enopt_config["linear_constraints"]["rhs_values"] = [1.0, -1.0, 1.0]
    results = BasicOptimizer(enopt_config, evaluator()).run().results
    assert results is None


def test_check_nonlinear_constraints(
    enopt_config: Any, evaluator: Any, test_functions: Any
) -> None:
    enopt_config["nonlinear_constraints"] = {
        "rhs_values": [0.0, 0.0, 0.0],
        "types": [ConstraintType.EQ, ConstraintType.LE, ConstraintType.GE],
        "scales": 10.0,
    }

    test_functions = (
        *test_functions,
        lambda variables, _: cast(NDArray[np.float64], variables[0]),
        lambda variables, _: cast(NDArray[np.float64], variables[0]),
        lambda variables, _: cast(NDArray[np.float64], variables[0]),
    )

    enopt_config["optimizer"]["max_functions"] = 1

    results = BasicOptimizer(enopt_config, evaluator(test_functions)).run().results
    assert results is not None

    enopt_config["nonlinear_constraints"]["rhs_values"] = [1.0, -1.0, 1.0]

    results = BasicOptimizer(enopt_config, evaluator(test_functions)).run().results
    assert results is None


def test_optimizer_variables_subset(enopt_config: Any, evaluator: Any) -> None:
    # Set the second variable a constant value, this will not affect the
    # optimization of the other variables in this particular test problem:
    enopt_config["variables"]["initial_values"] = [0.0, 1.0, 0.1]
    enopt_config["variables"]["indices"] = [0, 2]

    def assert_gradient(event: Event) -> None:
        assert event.results
        assert event.results is not None
        for item in event.results:
            if isinstance(item, GradientResults):
                assert item.gradients is not None
                assert item.gradients.weighted_objective[1] == 0.0
                assert np.all(item.gradients.objectives[:, 1] == 0.0)

    variables = (
        BasicOptimizer(enopt_config, evaluator())
        .add_observer(EventType.FINISHED_EVALUATION, assert_gradient)
        .run()
        .variables
    )
    assert variables is not None
    assert np.allclose(variables, [0.0, 1.0, 0.5], atol=0.02)


def test_optimizer_variables_subset_linear_constraints(
    enopt_config: Any, evaluator: Any
) -> None:
    # Set the second variable a constant value, this will not affect the
    # optimization of the other variables in this particular test problem:
    enopt_config["variables"]["initial_values"] = [0.0, 1.0, 0.1]
    # The second and third constraints are dropped because they involve
    # variables that are not optimized.
    enopt_config["linear_constraints"] = {
        "coefficients": [[1, 0, 1], [0, 1, 0], [1, 1, 1]],
        "rhs_values": [1.0, 1.0, 2.0],
        "types": [ConstraintType.EQ, ConstraintType.EQ, ConstraintType.EQ],
    }
    enopt_config["variables"]["indices"] = [0, 2]

    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.25, 1.0, 0.75], atol=0.02)


def test_parallelize(enopt_config: Any, evaluator: Any) -> None:
    enopt_config["optimizer"] = {
        "method": "differential_evolution",
        "max_iterations": 15,
        "options": {"seed": 123, "tol": 1e-10},
    }
    enopt_config["variables"]["lower_bounds"] = [0.15, 0.0, 0.0]
    enopt_config["variables"]["upper_bounds"] = [0.5, 0.5, 0.2]
    enopt_config["variables"]["initial_values"][0] = 0.2

    enopt_config["optimizer"]["parallel"] = False
    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.15, 0.0, 0.2], atol=3e-2)

    enopt_config["optimizer"]["parallel"] = True
    variables = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables is not None
    assert np.allclose(variables, [0.15, 0.0, 0.2], atol=3e-2)


def test_rng(enopt_config: Any, evaluator: Any) -> None:
    variables1 = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables1 is not None
    assert np.allclose(variables1, [0.0, 0.0, 0.5], atol=0.02)

    variables2 = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables2 is not None
    assert np.allclose(variables2, [0.0, 0.0, 0.5], atol=0.02)
    assert np.all(variables1 == variables2)

    enopt_config["gradient"]["seed"] = (1, DEFAULT_SEED)
    variables3 = BasicOptimizer(enopt_config, evaluator()).run().variables
    assert variables3 is not None
    assert np.allclose(variables3, [0.0, 0.0, 0.5], atol=0.02)
    assert not np.all(variables3 == variables1)



from functools import partial
from typing import Any, Dict, List

import numpy as np
import pytest
from numpy.typing import NDArray

from ropt.enums import ConstraintType, EventType
from ropt.plan import BasicOptimizer, Event
from ropt.plugins.realization_filter.default import (
    _get_cvar_weights_from_percentile,
    _sort_and_select,
)
from ropt.results import FunctionResults, GradientResults, Results


@pytest.fixture(name="enopt_config")
def enopt_config_fixture() -> Dict[str, Any]:
    return {
        "optimizer": {
            "speculative": True,
            "max_functions": 10,
        },
        "objective_functions": {
            "names": ["f1", "f2"],
            "weights": [0.75, 0.25],
        },
        "gradient": {
            "number_of_perturbations": 5,
            "perturbation_magnitudes": 0.01,
        },
        "realizations": {
            "weights": 5 * [1.0],
        },
        "variables": {
            "initial_values": 3 * [0],
        },
    }


@pytest.mark.parametrize(
    ("failed", "first", "last", "expected"),
    [
        ([False, False, False, False, False], 0, 1, [0, 0, 0, 0.4, 0.5]),
        ([False, False, False, False, False], 1, 2, [0, 0, 0.3, 0.4, 0]),
        ([False, False, False, True, False], 0, 1, [0, 0, 0.3, 0, 0.5]),
        ([False, False, True, False, False], 1, 2, [0.1, 0, 0.0, 0.4, 0]),
    ],
)
def test__sort_and_select(
    failed: List[bool], first: int, last: int, expected: List[float]
) -> None:
    values = np.array([3, 4, 2, 1, 0])
    weights = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
    result = _sort_and_select(values, weights, np.array(failed), first, last)
    assert np.allclose(result, expected)


@pytest.mark.parametrize(
    ("failed", "percentile", "weights"),
    [
        ([False, False, False, False, False], 0.1, [0.0, 0.0, 0.0, 0.0, 0.1]),
        ([False, False, False, False, False], 0.2, [0.0, 0.0, 0.0, 0.0, 0.2]),
        ([False, False, False, False, False], 0.3, [0.0, 0.0, 0.0, 0.1, 0.2]),
        ([False, False, False, False, False], 0.35, [0.0, 0.0, 0.0, 0.15, 0.2]),
        ([False, False, False, False, False], 0.4, [0.0, 0.0, 0.0, 0.2, 0.2]),
        ([False, False, False, False, False], 0.9, [0.2, 0.1, 0.2, 0.2, 0.2]),
        ([False, False, False, False, False], 1.0, [0.2, 0.2, 0.2, 0.2, 0.2]),
        ([False, False, False, True, False], 0.4, [0, 0, 0.15, 0.0, 0.25]),
    ],
)
def test__get_cvar_weights_from_percentile(
    failed: List[bool], percentile: float, weights: List[float]
) -> None:
    values = np.array([3, 4, 2, 1, 0])
    estimated_weights = _get_cvar_weights_from_percentile(
        values,
        np.array(failed),
        percentile,
    )
    assert np.allclose(estimated_weights, weights)


def _objective_function(
    variables: NDArray[np.float64],
    context: Any,
    target: NDArray[np.float64],
) -> float:
    diff: NDArray[np.float64] = variables - target
    # Make sure that some realizations yield a different result:
    if context.realization % 2 == 0:
        diff += 1.0
    result = np.sum(diff**2)
    # Make sure the other realizations have a worse result:
    if context.realization % 2 == 1:
        result += 10.0
    return float(result)


def _constraint_function(variables: NDArray[np.float64], context: Any) -> float:
    # Track how often this function is called.
    result = variables[0] + variables[2]
    # Break some realizations, same as in the distance function:
    if context.realization % 2 == 0:
        result = variables[0] + variables[2] - 10.0
    return float(result)


def _track_results(event: Event, result_list: List[Results]) -> None:
    assert event.results
    result_list.extend(event.results)


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_sort_filter_on_objectives(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    variables = BasicOptimizer(enopt_config, evaluator(functions)).run().variables
    assert variables is not None
    assert not np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)

    enopt_config["realization_filters"] = [
        {
            "method": "sort-objective",
            "options": {
                "sort": ["f1"],
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0]

    result_list: List[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .add_observer(
            EventType.FINISHED_EVALUATION,
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_sort_filter_on_objectives_with_constraints(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_constraint_function),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    enopt_config["nonlinear_constraints"] = {
        "rhs_values": 0.4,
        "types": [ConstraintType.LE],
    }
    enopt_config["realization_filters"] = [
        {
            "method": "sort-objective",
            "options": {
                "sort": [0],
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0]
    enopt_config["nonlinear_constraints"]["realization_filters"] = [0]
    result_list: List[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .add_observer(
            EventType.FINISHED_EVALUATION,
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [-0.05, 0.0, 0.45], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            assert result.evaluations.perturbed_constraints is not None
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_sort_filter_on_constraints(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_constraint_function),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations
    enopt_config["nonlinear_constraints"] = {
        "names": ["c1"],
        "rhs_values": 0.4,
        "types": [ConstraintType.LE],
    }
    enopt_config["realization_filters"] = [
        {
            "method": "sort-constraint",
            "options": {
                "sort": "c1",
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0]
    enopt_config["nonlinear_constraints"]["realization_filters"] = [0]
    result_list: List[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .add_observer(
            EventType.FINISHED_EVALUATION,
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [-0.05, 0.0, 0.45], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            assert result.evaluations.perturbed_constraints is not None
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_sort_filter_mixed(  # noqa: C901
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
    ]

    enopt_config["objective_functions"]["weights"] = [0.75, 0.25, 0.75, 0.25]
    enopt_config["objective_functions"]["names"] = ["f1", "f2", "f3", "f4"]
    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    objective_values: List[NDArray[np.float64]] = []

    def _add_objective(event: Event) -> None:
        if event.results:
            for item in event.results:
                if isinstance(item, FunctionResults):
                    assert item.functions is not None
                    objective_values.append(item.functions.weighted_objective)
        _track_results(event, result_list=result_list)

    # Apply the filtering to all objectives, giving the expected result.
    enopt_config["realization_filters"] = [
        {
            "method": "sort-objective",
            "options": {
                "sort": ["f1"],
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0, 0, 0]

    result_list: List[Results] = []
    results = (
        BasicOptimizer(enopt_config, evaluator(functions))
        .add_observer(EventType.FINISHED_EVALUATION, _add_objective)
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )

    # Apply filtering only to the first two, giving a wrong result.
    enopt_config["realization_filters"] = [
        {
            "method": "sort-objective",
            "options": {
                "sort": [0],
                "first": 3,
                "last": 4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0, -1, -1]

    result_list = []
    results = (
        BasicOptimizer(enopt_config, evaluator(functions))
        .add_observer(EventType.FINISHED_EVALUATION, _add_objective)
        .run()
        .results
    )
    assert results is not None
    assert not np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, :, :2] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, :, :2] != 0.0
                )
            assert np.all(
                result.evaluations.perturbed_objectives[filtered, :, 2:] != 0.0
            )

    # The first objective values should differ.
    assert objective_values[0] != objective_values[1]


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_cvar_filter_on_objectives(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    variables = BasicOptimizer(enopt_config, evaluator(functions)).run().variables
    assert variables is not None
    assert not np.allclose(variables, [0.0, 0.0, 0.5], atol=0.02)

    enopt_config["realization_filters"] = [
        {
            "method": "cvar-objective",
            "options": {
                "sort": [0],
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0]
    result_list: List[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .add_observer(
            EventType.FINISHED_EVALUATION,
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_cvar_filter_on_objectives_with_constraints(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_constraint_function),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    enopt_config["nonlinear_constraints"] = {
        "rhs_values": 0.4,
        "types": [ConstraintType.LE],
    }
    enopt_config["realization_filters"] = [
        {
            "method": "cvar-objective",
            "options": {
                "sort": [0],
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0]
    enopt_config["nonlinear_constraints"]["realization_filters"] = [0]
    result_list: List[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .add_observer(
            EventType.FINISHED_EVALUATION,
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [-0.05, 0.0, 0.45], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            assert result.evaluations.perturbed_constraints is not None
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_cvar_filter_on_constraints(
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_constraint_function),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations

    enopt_config["nonlinear_constraints"] = {
        "rhs_values": 0.4,
        "types": [ConstraintType.LE],
    }
    enopt_config["realization_filters"] = [
        {
            "method": "cvar-constraint",
            "options": {
                "sort": 0,
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0]
    enopt_config["nonlinear_constraints"]["realization_filters"] = [0]
    result_list: List[Results] = []
    results = (
        BasicOptimizer(
            enopt_config,
            evaluator(functions),
        )
        .add_observer(
            EventType.FINISHED_EVALUATION,
            partial(_track_results, result_list=result_list),
        )
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [-0.05, 0.0, 0.45], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            assert result.evaluations.perturbed_constraints is not None
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )
                assert np.all(
                    result.evaluations.perturbed_constraints[filtered, ...] != 0.0
                )


@pytest.mark.parametrize("split_evaluations", [True, False])
def test_cvar_filter_mixed(  # noqa: C901
    enopt_config: Any,
    evaluator: Any,
    split_evaluations: bool,
) -> None:
    functions = [
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
        partial(_objective_function, target=np.array([0.5, 0.5, 0.5])),
        partial(_objective_function, target=np.array([-1.5, -1.5, 0.5])),
    ]

    enopt_config["optimizer"]["split_evaluations"] = split_evaluations
    enopt_config["objective_functions"]["weights"] = [0.75, 0.25, 0.75, 0.25]
    enopt_config["objective_functions"]["names"] = ["f1", "f2", "f3", "f4"]

    objective_values: List[NDArray[np.float64]] = []

    def _add_objective(event: Event) -> None:
        if event.results is not None:
            for item in event.results:
                if isinstance(item, FunctionResults):
                    assert item.functions is not None
                    objective_values.append(item.functions.weighted_objective)
        _track_results(event, result_list=result_list)

    # Apply the filtering to all objectives, giving the expected result.
    enopt_config["realization_filters"] = [
        {
            "method": "cvar-objective",
            "options": {
                "sort": ["f1"],
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0, 0, 0]

    result_list: List[Results] = []
    results = (
        BasicOptimizer(enopt_config, evaluator(functions))
        .add_observer(EventType.FINISHED_EVALUATION, _add_objective)
        .run()
        .results
    )
    assert results is not None
    assert np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, ...] != 0.0
                )

    # Apply filtering only to the first two, giving a wrong result.
    enopt_config["realization_filters"] = [
        {
            "method": "cvar-objective",
            "options": {
                "sort": [0],
                "percentile": 0.4,
            },
        },
    ]
    enopt_config["objective_functions"]["realization_filters"] = [0, 0, -1, -1]

    result_list = []
    results = (
        BasicOptimizer(enopt_config, evaluator(functions))
        .add_observer(EventType.FINISHED_EVALUATION, _add_objective)
        .run()
        .results
    )
    assert results is not None
    assert not np.allclose(results.evaluations.variables, [0.0, 0.0, 0.5], atol=0.02)
    for result in result_list:
        assert result is not None
        if isinstance(result, FunctionResults):
            assert result.realizations is not None
            assert result.realizations.objective_weights is not None
            filtered = result.realizations.objective_weights[0, :] == 0.0
        if isinstance(result, GradientResults):
            if split_evaluations:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, :, :2] == 0.0
                )
            else:
                assert np.all(
                    result.evaluations.perturbed_objectives[filtered, :, :2] != 0.0
                )
            assert np.all(
                result.evaluations.perturbed_objectives[filtered, :, 2:] != 0.0
            )

    # The first objective values should differ.
    assert objective_values[0] != objective_values[1]



"""This module defines the abstract base class for plugins."""

from __future__ import annotations

from abc import ABC, abstractmethod


class Plugin(ABC):
    """Abstract base class for plugins.

    `ropt`  plugins should derive from this base class, which specifies an
    `is_supported` method.
    """

    @abstractmethod
    def is_supported(self, method: str, *, explicit: bool) -> bool:
        """Check whether a given method is supported.

        This method is called by the
        [`is_supported`][ropt.plugins.PluginManager.is_supported] method of
        [`PluginManager`][ropt.plugins.PluginManager] objects to verify if a
        specific method is supported by this plugin.

        If the `explicit` flag is set to `True`, the plugin has been explicitly
        requested. In this case, `True` should be returned if the specified
        optimization method is supported.

        If `explicit` is `False`, the plugin manager is performing a general
        search across all available plugins for the requested method. In this
        scenario, `True` should only be returned if the plugin's method is
        compatible with this search. If the search compatibility is not
        supported, then `False` should be returned, even if the plugin
        technically supports the requested method.

        For example, the
        [`external`][ropt.plugins.optimizer.external.ExternalOptimizer]
        optimizer plugin does not define its own methods but launches methods
        from other plugins as an external process. Therefore, the `external`
        optimizer plugin must always be specified explicitly. As a result, its
        `is_supported` method returns `False` if `explicit` is `False`.

        Args:
            method:   The name of the method to check.
            explicit: Indicates whether the plugin was explicitly requested.

        Returns:
            True if the method is supported; otherwise, False.
        """



"""Example of using the ScriptOptimizer class."""

import argparse
import json
import sys
from pathlib import Path
from typing import Tuple


def rosenbrock(x: float, y: float, realization: int, coefficients_path: Path) -> float:
    """Rosenbrock function."""
    with coefficients_path.open("r", encoding="utf-8") as f:
        coefficients = json.load(f)
    a, b = coefficients[realization]
    return float((a - x) ** 2 + b * (y - x * x) ** 2)


def _read_point(filename: Path) -> Tuple[float, float]:
    with filename.open("r", encoding="utf-8") as f:
        variables = json.load(f)
    return variables["x"]["1"], variables["y"]["1"]


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--vars", type=Path)
    arg_parser.add_argument("--coefficients", type=Path)
    arg_parser.add_argument("--realization", type=str)
    arg_parser.add_argument("--out", type=Path)
    options, _ = arg_parser.parse_known_args(args=sys.argv[1:])

    x, y = _read_point(options.vars)

    value = rosenbrock(x, y, options.realization, options.coefficients)

    with options.out.open("w", encoding="utf-8") as f:
        f.write(f"{value:.15e}")



"""This module implements the default tracker handler."""

from __future__ import annotations

import sys
from copy import deepcopy
from typing import TYPE_CHECKING, Literal, Optional, Tuple, Union

from pydantic import BaseModel, ConfigDict, Field

from ropt.config.validated_types import ItemOrSet  # noqa: TCH001
from ropt.enums import EventType
from ropt.plugins.plan.base import ResultHandler

from ._utils import _get_all_results, _get_last_result, _update_optimal_result

if sys.version_info >= (3, 11):
    pass
else:
    pass


if TYPE_CHECKING:
    from ropt.config.plan import ResultHandlerConfig
    from ropt.plan import Event, Plan
    from ropt.results import FunctionResults


class DefaultTrackerHandler(ResultHandler):
    """The default tracker results handler object.

    This handler tracks the
    [`Results`][ropt.results.Results] objects that it receives and selects one
    to retain in a variable. Currently it tracks either the last result it
    receives, or the best result. The best result is defined as the result that
    has the lowest weighted objective value. Optionally, results may be filtered
    by checking for violations of constraints, by comparing constraint values to
    a threshold.

    The tracker step uses the [`DefaultTrackerHandlerWith`]
    [ropt.plugins.plan._tracker.DefaultTrackerHandler.DefaultTrackerHandlerWith]
    configuration class to parse the `with` field of the
    [`ResultHandler`][ropt.config.plan.ResultHandlerConfig] used to specify this
    handler in a plan configuration.
    """

    class DefaultTrackerHandlerWith(BaseModel):
        """Parameters for the tracker results handler.

        The tracker stores the tracked results in the variable specified by the
        `var` field. The `type` parameter determines which result is tracked:

        - `"best"`: Tracks the best result added.
        - `"last"`: Tracks the last result added.
        - `"all"`:  Store a tuple with all results.

        If `constraint_tolerance` is set, results that exceed this tolerance on
        constraint values are not tracked.

        The `tags` field allows optional labels to be attached to each result,
        assisting result handlers in filtering relevant results.

        Attributes:
            var:                  The name of the variable to store the tracked result.
            tags:                 Tags to filter the sources to track.
            type:                 Specifies the type of result to store.
            constraint_tolerance: An optional constraint tolerance level.
        """

        var: str
        tags: ItemOrSet[str]
        type_: Literal["best", "last", "all"] = Field(default="best", alias="type")
        constraint_tolerance: Optional[float] = 1e-10

        model_config = ConfigDict(
            extra="forbid",
            validate_default=True,
            frozen=True,
        )

    def __init__(self, config: ResultHandlerConfig, plan: Plan) -> None:
        """Initialize a default tracker results handler object.

        Args:
            config: The configuration of the step.
            plan:   The plan that runs this step.
        """
        super().__init__(config, plan)
        self._with = self.DefaultTrackerHandlerWith.model_validate(config.with_)
        self.plan[self._with.var] = None

    def handle_event(self, event: Event) -> Event:
        """Handle an event.

        Args:
            event: The event to handle.

        Returns:
            The (possibly modified) event.
        """
        if (
            event.event_type
            in {
                EventType.FINISHED_EVALUATION,
                EventType.FINISHED_EVALUATOR_STEP,
            }
            and event.results is not None
            and (event.tags & self._with.tags)
        ):
            results: Union[Optional[FunctionResults], Tuple[FunctionResults, ...]]
            results = None
            if self._with.type_ == "all":
                results = _get_all_results(
                    event.results, self._with.constraint_tolerance
                )
                self.plan[self._with.var] = deepcopy(results)
            elif self._with.type_ == "best":
                results = _update_optimal_result(
                    self.plan[self._with.var],
                    event.results,
                    self._with.constraint_tolerance,
                )
            elif self._with.type_ == "last":
                results = _get_last_result(
                    event.results, self._with.constraint_tolerance
                )
            if results is not None:
                self.plan[self._with.var] = deepcopy(results)
        return event



"""This module defines the base classes for optimization plan plugins.

Optimization plan steps and result handlers can be added via the plugin
mechanism to implement additional functionality. This is done by creating a
plugin class that derives from the
[`PlanPlugin`][ropt.plugins.plan.base.PlanPlugin] class. It needs to define a
[`create`][ropt.plugins.plan.base.PlanPlugin.create] method that generates the
plan objects.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Union

from ropt.plugins.base import Plugin

if TYPE_CHECKING:
    from ropt.config.plan import PlanStepConfig, ResultHandlerConfig
    from ropt.plan import Event, OptimizerContext, Plan


class PlanPlugin(Plugin):
    """Abstract base class for plan plugins.

    This base class serves as the foundation for all plugins used in an
    optimization plan, including plan steps and result handlers. Any plugin
    derived from this class can be built-in, installed via a plugin mechanism,
    or dynamically loaded. During the execution of an optimization plan, the
    required plugin is located through the
    [`PluginManager`][ropt.plugins.PluginManager], which then uses the plugin's
    `create` function to instantiate either a
    [`PlanStep`][ropt.plugins.plan.base.PlanStep] or a
    [`ResultHandler`][ropt.plugins.plan.base.ResultHandler] as needed.

    Note: Plan functions
        A plan plugin can also provide functions to be used by the plan`s
        expression evaluator. To achieve this, the plugin should override the
        `functions` property of the
        [`PlanPlugin`][ropt.plugins.plan.base.PlanPlugin] class. This property
        should return a dictionary that maps function names to callables, which
        are subsequently added to the expression evaluator.
    """

    @abstractmethod
    def create(
        self,
        config: Union[ResultHandlerConfig, PlanStepConfig],
        context: OptimizerContext,
    ) -> Union[ResultHandler, PlanStep]:
        """Create a step or result handler.

        This factory function instantiates either a step or a result handler
        object based on the provided configuration. The configuration determines
        which type of object to return—either a step or a result handler.

        Args:
            config:  The configuration for the plan object.
            context: The context in which the plan operates.
        """

    @property
    def functions(self) -> Dict[str, Any]:
        """Return plan functions implemented by the plugin.

        Returns:
            The functions.
        """
        return {}


class PlanStep(ABC):
    """Base class for steps.

    steps loaded and created by the plugin manager must inherit from this
    abstract base class. It defines a `run` method that must be overridden to
    implement the specific functionality of each step. The class also provides
    default properties for accessing the configuration and the plan that
    executes the step.
    """

    def __init__(self, config: PlanStepConfig, plan: Plan) -> None:
        """Initialize the step.

        The `config` and `plan` arguments are accessible as `step_config`
        and `plan` properties.

        Args:
            config: The configuration for this step.
            plan:   The parent plan that manages this step.
        """
        self._step_config = config
        self._plan = plan

    @property
    def step_config(self) -> PlanStepConfig:
        """Return the step object's configuration.

        Returns:
            The configuration object.
        """
        return self._step_config

    @property
    def plan(self) -> Plan:
        """Return the plan that executes the step.

        Returns:
            The plan object.
        """
        return self._plan

    @abstractmethod
    def run(self) -> None:
        """Execute the step object.

        This method must be overloaded to implement the functionality of the
        step.
        """


class ResultHandler(ABC):
    """Base class for result handler objects.

    Result handlers loaded and created by the plugin manager must inherit from
    this abstract base class. It defines a `handle_event` method that must be
    overridden to implement specific functionality and provides default
    properties for accessing the handler's configuration and the plan that runs
    it.
    """

    def __init__(self, config: ResultHandlerConfig, plan: Plan) -> None:
        """Initialize a results handler.

        The `config` and `plan` arguments are accessible as `handler_config`
        and `plan` properties.

        Args:
            config: The configuration of the handler object.
            plan:   The parent plan that contains the object.
        """
        self._handler_config = config
        self._plan = plan

    @property
    def handler_config(self) -> ResultHandlerConfig:
        """Return the configuration of the handler object.

        Returns:
            The configuration object.
        """
        return self._handler_config

    @property
    def plan(self) -> Plan:
        """Return the associated plan that executes the handler.

        Returns:
            The plan object.
        """
        return self._plan

    @abstractmethod
    def handle_event(self, event: Event) -> Event:
        """Handle and propagate an event.

        This method must be overloaded to implement the functionality of the
        handler, based on the information passed via the event object. The
        handler may modify the event, which may be propagated further to other
        handlers or to callbacks that react to the event.

        Returns:
            The event, possibly modified.
        """



"""This module implements the default table handler."""

from __future__ import annotations

import sys
from pathlib import Path  # noqa: TCH003
from typing import TYPE_CHECKING, Dict, Literal, Optional

from pydantic import BaseModel, ConfigDict

from ropt.config.validated_types import ItemOrSet  # noqa: TCH001
from ropt.enums import EventType
from ropt.plugins.plan.base import ResultHandler
from ropt.report import ResultsTable
from ropt.results import convert_to_maximize

if sys.version_info >= (3, 11):
    pass
else:
    pass


if TYPE_CHECKING:
    from ropt.config.plan import ResultHandlerConfig
    from ropt.plan import Event, Plan


class DefaultTableHandler(ResultHandler):
    """The default table results handler object.

    The table results handler generates a table that summarizes the results it
    receives. This is implemented through a
    [`ResultsTable`][ropt.report.ResultsTable] object; refer to its
    documentation for detailed information on table generation and formatting.

    This table handler uses the [`DefaultTableHandlerWith`]
    [ropt.plugins.plan._table.DefaultTableHandler.DefaultTableHandlerWith]
    configuration class to parse the `with` field in the
    [`ResultHandler`][ropt.config.plan.ResultHandlerConfig] used to specify this
    handler in a plan configuration.
    """

    class DefaultTableHandlerWith(BaseModel):
        """Parameters for the table results handler.

        The `columns`, `path`, `table_type`, and `min_header_len` fields
        correspond to the parameters of the
        [`ResultsTable`][ropt.report.ResultsTable] constructor. Refer to its
        documentation for more detailed information on these arguments.

        The `maximize` field allows adaptation for maximization problems.
        Typically, results are generated for minimization problems, but setting
        `maximize` to `True` will invert this interpretation for scenarios where
        the negative of an objective function is minimized.

        The `tags` field enables optional labels to be attached to results,
        which assists in filtering relevant results for the table.

        Attributes:
            tags:           Tags of the sources to track.
            columns:        A mapping of column names for the results table.
            path:           The file path where the results table will be saved.
            table_type:     The format or type of table to generate.
            min_header_len: The minimum number of header lines to include.
            maximize:       If `True`, interprets results as a maximization
                            problem rather than the default minimization.
        """

        tags: ItemOrSet[str]
        columns: Dict[str, str]
        path: Path
        table_type: Literal["functions", "gradients"] = "functions"
        min_header_len: Optional[int] = None
        maximize: bool = False

        model_config = ConfigDict(
            extra="forbid",
            validate_default=True,
            frozen=True,
        )

    def __init__(self, config: ResultHandlerConfig, plan: Plan) -> None:
        """Initialize a default table results handler object.

        Args:
            config: The configuration of the step.
            plan:   The plan that runs this step.
        """
        super().__init__(config, plan)
        self._with = self.DefaultTableHandlerWith.model_validate(config.with_)
        self._table = ResultsTable(
            self._with.columns,
            self._with.path,
            table_type=self._with.table_type,
            min_header_len=self._with.min_header_len,
        )

    def handle_event(self, event: Event) -> Event:
        """Handle an event.

        Args:
            event: The event to handle.

        Returns:
            The (possibly modified) event.
        """
        if (
            event.event_type
            in {
                EventType.FINISHED_EVALUATION,
                EventType.FINISHED_EVALUATOR_STEP,
            }
            and event.results is not None
            and (event.tags & self._with.tags)
        ):
            self._table.add_results(
                event.config,
                (
                    (convert_to_maximize(item) for item in event.results)
                    if self._with.maximize
                    else event.results
                ),
            )
        return event



"""This module defines the abstract base class for optimization steps."""

from __future__ import annotations

from typing import Optional, Tuple

import numpy as np

from ropt.results import FunctionResults, Results


def _get_new_optimal_result(
    optimal_result: Optional[FunctionResults], results: FunctionResults
) -> Optional[FunctionResults]:
    if optimal_result is None:
        return results
    assert optimal_result.functions is not None
    assert results.functions is not None
    optimal = optimal_result.functions.weighted_objective
    objective = results.functions.weighted_objective
    if objective < optimal:
        return results
    return None


def _check_bound_constraints(
    results: FunctionResults, constraint_tolerance: float
) -> bool:
    if results.bound_constraints is not None:
        lower_violations = results.bound_constraints.scaled_lower_violations
        if lower_violations is None:
            lower_violations = results.bound_constraints.lower_violations
        upper_violations = results.bound_constraints.scaled_upper_violations
        if upper_violations is None:
            upper_violations = results.bound_constraints.upper_violations
        for violations in (lower_violations, upper_violations):
            if (
                violations is not None
                and np.any(violations > constraint_tolerance).item()
            ):
                return False
    return True


def _check_constraints(
    results: FunctionResults, constraint_tolerance: Optional[float]
) -> bool:
    if constraint_tolerance is None:
        return True

    if not _check_bound_constraints(results, constraint_tolerance):
        return False

    if results.linear_constraints is not None and (
        results.linear_constraints.violations is not None
        and np.any(results.linear_constraints.violations > constraint_tolerance).item()
    ):
        return False

    if results.nonlinear_constraints is not None:
        violations = results.nonlinear_constraints.scaled_violations
        if violations is None:
            violations = results.nonlinear_constraints.violations
        if violations is not None and np.any(violations > constraint_tolerance).item():
            return False

    return True


def _get_last_result(
    results: Tuple[Results, ...], constraint_tolerance: Optional[float]
) -> Optional[FunctionResults]:
    return next(
        (
            item
            for item in reversed(results)
            if (
                isinstance(item, FunctionResults)
                and item.functions is not None
                and _check_constraints(item, constraint_tolerance)
            )
        ),
        None,
    )


def _update_optimal_result(
    optimal_result: Optional[FunctionResults],
    results: Tuple[Results, ...],
    constraint_tolerance: Optional[float],
) -> Optional[FunctionResults]:
    return_result: Optional[FunctionResults] = None
    for item in results:
        if (
            isinstance(item, FunctionResults)
            and item.functions is not None
            and _check_constraints(item, constraint_tolerance)
        ):
            new_optimal_result = _get_new_optimal_result(optimal_result, item)
            if new_optimal_result is not None:
                optimal_result = new_optimal_result
                return_result = new_optimal_result
    return return_result


def _get_all_results(
    results: Tuple[Results, ...], constraint_tolerance: Optional[float]
) -> Tuple[FunctionResults, ...]:
    return tuple(
        item
        for item in results
        if (
            isinstance(item, FunctionResults)
            and item.functions is not None
            and _check_constraints(item, constraint_tolerance)
        )
    )
